import collections
from datetime import datetime
import logging
import re
import sys

from django.conf import settings
from django.contrib.auth.decorators import login_required
from django.db import transaction
from django.http import HttpResponse
from django.shortcuts import redirect, render
from django.utils import simplejson
from django.utils.timezone import get_default_timezone
from django.views.decorators.csrf import csrf_exempt

from fa.apps import FoneAstraApp
from fa.models import IncomingMessage, OutgoingMessage
from fa.transports.base import TransportBase, NoTransportException
from fa.transports.smssync import Transport


logger = logging.getLogger(__name__)


@login_required
def index(req):
    apps = FoneAstraApp.find_all_fa_apps()
    if len(apps) == 1:
        return redirect(apps[0].app_index)
    return render(req, "fa/index.html")

@csrf_exempt
def smssync(request):
    if 'secret' not in request.REQUEST:
        if settings.DEBUG:
            return HttpResponse(simplejson.dumps(_debug()), mimetype="application/json")
        else:
            return HttpResponse("Debug mode currently disabled.")
    with transaction.commit_on_success():
        payload = {
            'secret': request.REQUEST['secret'],
        }
        transport = find_transport(payload['secret'])

        if request.method == 'POST':
            sender = request.POST['from']
            content = request.POST['message']

            # See that whole "hour = blah % 24" thing?  We need to handle
            # all of this weird regex stuff instead of just using strptime
            # because SMSSync returns times between midnight and one as
            # 24:XX, not 00:XX.
            matches = re.match(
                "(\d+)-(\d+)-(\d+) (\d+):(\d+)",
                request.POST['sent_timestamp'],
            )
            month = int(matches.group(1))
            day = int(matches.group(2))
            year = int(matches.group(3)) + 2000
            hour = int(matches.group(4)) % 24
            minute = int(matches.group(5))

            send_time = datetime(year, month, day, hour, minute).replace(
                tzinfo=get_default_timezone()
            )
            message = IncomingMessage(
                transport=transport.name,
                sender=sender,
                content=content,
                send_time=send_time,
            )
            payload['success'] = ("true" if TransportBase.receive(message) else "false")

        outgoing_messages = OutgoingMessage.objects.filter(
            transport=transport.name
        )
        if len(outgoing_messages) > 0:
            payload['task'] = "send"
            messages = [{
                "to": msg.target,
                "message": msg.content,
            } for msg in outgoing_messages]

            # Don't delete if we're viewing this in a browser.
            if ('HTTP_USER_AGENT' not in request.META
                    or 'SMSSync' in request.META['HTTP_USER_AGENT']):
                outgoing_messages.delete()
            payload['messages'] = messages

        reply = {
            "payload": payload
        }
        return HttpResponse(simplejson.dumps(reply), mimetype="application/json")

def find_transport(secret):
    """
    Finds the SMSSync transport with the given secret.  Throws an exception if
    if it cannot find one.

    """
    for transport in TransportBase.get_all():
        if isinstance(transport, Transport):
            if transport.config['secret'] == secret:
                return transport
    raise NoTransportException()


def _debug():
    """
    Returns debugging information in the form of of a dict mapping SMSSync
    secrets to list of all messages using that transport.

    """
    messages = collections.defaultdict(list)
    for msg in OutgoingMessage.objects.all():
        messages[msg.transport].append(str(msg))
    return messages



from collections import defaultdict
import logging

from django.conf import settings
from django.utils.importlib import import_module

from fa.apps import FoneAstraApp


logger = logging.getLogger(__name__)


class TransportBase(object):
    """
    Base class for all transports.  Defines methods that transports must
    implement.

    All Transport constructors must take the name of the transport (string) as
    their first argument, and they are passed the entire configuration dict
    (from the global settings file) as **kwargs.

    """

    # Lazily initialize our transports list.
    _transports = None

    @classmethod
    def get_transport(cls, name):
        """Gets a transport object from its configured name."""
        cls._generate_transports()
        if name not in cls._transports:
            raise NoTransportException("Could not find transport " + str(name))
        return cls._transports[name]
    
    @classmethod
    def get_all(cls):
        """Returns a list of all transports."""
        cls._generate_transports()
        return [b for a, b in cls._transports.items()]

    @classmethod
    def _generate_transports(cls):
        """Populates the _transports field."""
        if cls._transports is not None:
            return

        transports_local = {}
        for name, config in settings.TRANSPORTS.iteritems():
            transport_class = import_module(config["ENGINE"]).Transport
            transports_local[name] = transport_class(name, config)
            logger.debug("Adding transport {name}".format(name=name))
        cls._transports = transports_local

    @classmethod
    def receive(cls, msg):
        """
        Call this function to process an incoming message from a FoneAstra device.
        Returns true if an app or device handled the message, False otherwise.

        Arguments:
            msg - an IncomingMessage

        """
        
        # First check to see if this is a message from a registered device.
        for app in FoneAstraApp.find_all_fa_apps():
            device = app.get_device(msg.sender)
            if device:
                logger.info('Received message from FA device {name}: {content}'.format(
                    name=device.name,
                    content=msg.content,
                ))
                if msg.transport != device.transport:
                    logger.warning('Received message from FA device {name} on transport {transport} (expected {expected})'.format(
                        name=device.name,
                        transport=msg.transport,
                        expected=device.transport,
                    ))
                return device.receive(msg)

        # If it's not a registered device, just have every app look at it in order.
        for app in FoneAstraApp.find_all_fa_apps():
            if app.receive(msg):
                logger.info('Message {content} from {sender} successfully processed by {app}'.format(
                    content=msg.content,
                    sender=msg.sender,
                    app=app.app_name,
                ))
                return True

        # None of the apps want it.
        logger.warning('Unprocessed message from {sender}: {content}'.format(
            sender=msg.sender,
            content=msg.content
        ))
        return False

    @classmethod
    def send(cls, msg):
        """
        Sends a single message consisting of "content" to the target.

        Arguments:
            msg - OutgoingMessage

        """
        transport = cls.get_transport(msg.transport)
        transport._send(msg)

    @classmethod
    def send_batch(self, lst):
        """
        Sends all of the messages in lst, since some transports have more
        efficient options for batch sending.

        Arguments:
            lst - a list of OutgoingMessages

        """
        messages = defaultdict(list)
        for msg in lst:
            messages[msg.transport].append(msg)
        for transport_name, msg_list in messages.iteritems():
            transport = cls.get_transport(transport_name)
            transport._send_batch(msg_list)

    def _send(self, msg):
        """
        Private method to hand the actual sending.

        """
        pass

    def _send_batch(self, lst):
        """
        Private method to handle the actual batch sending.

        """
        for msg in lst:
            self._send(msg)

class NoTransportException(Exception):
    """Raised when attempting to send a message with an invalid transport."""
    def __init__(self, value=""):
        self.value=value
    def __str__(self):
        return repr(self.value)



from fa.transports.base import TransportBase

class Transport(TransportBase):
    """
    Transport for using an Android phone running SMSSync as the router.

    Configuration options:
        "secret" - SMSSync secret to use for device verification (no two SMSSync
                   transports running on the same server may have the same
                   secret)
    """


    def __init__(self, name, config):
        self.name = name
        self.config = config

    def _send(self, msg):
        """
        These messages get cleared in the HTTP thread, so just save them.

        """
        msg.save()



import inspect

from django.conf import settings
from django.core.exceptions import ObjectDoesNotExist

class FoneAstraApp(object):
    """
    Core app that all FoneAstra apps should subclass.  Includes extra methods
    and fields for interfacing with FoneAstra, and FoneAstra autodiscovery
    only includes apps that subcless FoneAstraApp.

    """

    # The name of the app (string).
    app_name = None

    # The app's index view name (string).
    # Example: `app name`_index
    app_index = None

    # The class definitions of all FoneAstra devices in this app.
    device_types = ()

    # Used to lazily initialize the apps list.
    _apps = None

    @classmethod
    def find_all_fa_apps(cls):
        if cls._apps is not None:
            return cls._apps
        for app_name in settings.INSTALLED_APPS:
            try:
                __import__(app_name + ".app")
            except ImportError:
                pass
        cls._apps = [x for x in cls.__subclasses__()
            if inspect.getmodule(x).__package__ in settings.INSTALLED_APPS]
        return cls._apps

    @classmethod
    def get_device(cls, phone_number):
        """
        Returns the FoneAstra device corresponding to this project and
        phone number, or None if no such device exists.

        """
        for device_type in cls.device_types:
            try:
                return device_type.objects.get(phone_number=phone_number)
            except ObjectDoesNotExist:
                pass
        return None

    @classmethod
    def receive(cls, msg):
        """
        This method gets called if a message arrives that did not come from any
        registered FoneAstra device in any app.  This method is most frequently
        called for configuration messages.

        """
        pass



import logging, logging.handlers
import threading

from django.db import models
from django.conf import settings

from fa.transports.base import TransportBase
from fa.utils import regex_test


class FoneAstraDevice(models.Model):
    """
    Representation of a FoneAstra device.  All devices in FoneAstra apps should
    subclass this class.
    """

    class Meta:
        abstract = True
        unique_together = ('transport', 'name')

    # YES I KNOW HOW UN-DJANGONIC THIS IS.
    TRANSPORT_CHOICES = [(x.name, x.name) for x in TransportBase.get_all()]

    # The name of this device.
    name = models.CharField(blank=True, max_length=200)

    # The phone number of this device.
    phone_number = models.CharField(max_length=25)

    # The transport to use to talk to this device.
    transport = models.CharField(blank=True, null=True, max_length=255,
        choices=TRANSPORT_CHOICES)

    # The receivers for this device.  A tuple of tuples, where each inner tuple
    # contains a regex and a function.
    # Example: receivers = (
    #              (handlera.pattern, handlera.func),
    #              (handlerb.pattern, handlerb.func),
    #          )
    # Receivers are called with the following arguments:
    #   device - the FoneAstra device
    #   send_time - the time the message was sent
    #   *args - the matching groups of the pattern.
    receivers = ()

    def __unicode__(self):
        return "{cls}: {name}".format(
            cls=str(self.__class__.__name__),
            name=self.name,
        )

    def receive(self, msg):
        """
        This is called when a message is received from this device.

        Arguments:
            msg - the IncomingMessage corresponding to this device

        """
        func, match = regex_test(self.receivers, msg.content)
        if func:
            return func(self, msg.send_time, *match.groups())
        return False

    def send(self, message):
        """
        Sends the given message to this FoneAstra device using this device's
        configured transport. Returns True if the message was sent successfully,
        but makes no guarantees that it was received.

        Arguments:
            message - string
        """
        msg = OutgoingMessage(
            target = self.phone_number,
            content = message,
            transport = self.transport,
        )
        TransportBase.send(msg)


class IncomingMessage(models.Model):
    """
    Representation of an incoming message, such as that from a device.
    """

    sender = models.CharField(max_length=25)

    content = models.CharField(max_length=255)

    send_time = models.DateTimeField()

    transport = models.CharField(max_length=255)

    def __unicode__(self):
        return ("{sender}: {content} !@ {send_time}".format(
            sender=self.sender,
            content=self.content,
            send_time=self.send_time
        ))


class OutgoingMessage(models.Model):
    """
    Representation of an outgoing message, such as an alarm.
    """

    target = models.CharField(max_length=20)

    content = models.CharField(max_length=255)

    transport = models.CharField(max_length=255)

    def __unicode__(self):
        return ("{target}: {content}".format(
            target=self.target,
            content=self.content,
        ))



#!/usr/bin/env python
# vim: ai ts=4 sts=4 et sw=4 coding=utf-8

import decimal
import logging
import re

from django.core.exceptions import ObjectDoesNotExist
from fa.models import *


logger = logging.getLogger(__name__)


def convert_temp(chars):
    """Unpacks temperatures from their packed forms. Each temperature consists
    of a 12-bit two's complement number, with 1 sign bit, 7 integer bits,
    and 4 fraction bits, and is packed into two 6-bit ASCII chars.

    Arguments:
        chars   -   string of length >= 2
    """
    # Encoder string of allowable ascii chars
    encoder = "0123456789ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz+/"

    # convert temp reading from ascii to decimal
    upper_ascii_char = chars[0]
    lower_ascii_char = chars[1]
    upper_int = encoder.index(upper_ascii_char)
    lower_int = encoder.index(lower_ascii_char)
    upper_bits = bin(upper_int)[2:].zfill(6)
    lower_bits = bin(lower_int)[2:].zfill(6)
    sign_bit = upper_bits[0]
    if (int(sign_bit) == 1):
        sign = "-"
    else:
        sign = "+"
    int_bits = upper_bits[1:6] + lower_bits[0:2]
    dec_bits = lower_bits[2:6]
    if (int(sign_bit) == 1):
        integer_num = int(int_bits,2) ^ int("1111111",2)
    else:
        integer_num = int(int_bits, 2)
    decimal_num = float(int(dec_bits, 2))
    if (int(sign_bit) == 1):
        fractional_dec = ((16.0 - decimal_num) / 16.0)
    else:
        fractional_dec = (decimal_num / 16.0)
    complete_num = integer_num + fractional_dec
    temp_str = "%s%.1f" % (sign, complete_num)
    temp = decimal.Decimal(temp_str)
    return temp


# This next function is copy/pasted from the old Reporters class
def field_bundles(qd, *keys):
    """When an HTML form has multiple fields with the same name, they are received
       as a single key, with an array of values -- which looks something like:

     { "backend": ["a", "b"], "identity": ["111", "222"], **other_stuff }

     ...which isn't very easy to iterate as pairs of "backend" and "identity";
     which is most often what we want to do. This method inverts the axis,
     and returns the same data in the less verbose, and more useful format:

     [("a", "111"), ("b", "222")]

     The order of tuples is preserved from the querydict - this USUALLY means
     that they'll be output in the same order as the controls on the HTML form
     that the request was triggered by - but that's not guaranteed. The order
     of the values inside the tuples (originally, the keys) is derrived from
     the *keys argument(s). Using the input from the first example, to output
     the second, in a familiar view:

     field_bundles(request.POST, "backend", "identity")"""

    bundles = []
    length = None

    # check that all values are of the same
    # length, even if that is zero, to avoid
    # creating half-bundles with missing data
    for key in keys:

        # if this is the first pair, then store the length
        # of the value, to check against subsequent pairs
        kl = len(qd.getlist(key))
        if length is None:
            length = kl

        # not the first pair, so we know long the value
        # SHOULD be - if it isn't exactly right, abort
        elif length != kl:
            raise IndexError(
                "for %s, expected length of %d, got %d" % (
                    key, length, kl))

    # iterate all of the data that we're going to
    # capture, and invert the axis, to group bundles
    # togeter, rather than parameters
    for n in range(0, length):
        bundles.append([qd.getlist(k)[n] for k in keys])

    return bundles


def chunks(iterable, n):
    """Yield successive n-sized chunks from iterable. Stolen from Ned
    Batchelder on StackOverflow.

    """
    for i in range(0, len(iterable), n):
        yield iterable[i:i+n]


def as_int(val):
    """Returns the value of the argument in hexadecimal.  Used so that
    we can abstract which base to use.

    """
    return int(val, 16)


def get_temp_sequence(readings):
    """Take in a string containing encoded temperatures and return a
    list of decimals.

    """
    return [convert_temp(x) for x in chunks(readings, 2)]


def regex_test(regex_list, input_string):
    """Given a list of (regex, object) tuples and an input string, matches
    the input string against each regex in turn and returns a tuple containing
    the object and the regex match object if the string matches. Only returns
    the first match.

    """
    for regex, obj in regex_list:
        match_obj = re.match(regex, input_string, re.IGNORECASE)
        if match_obj:
            logger.debug("Chose match {regex} for {input}".format(
                regex=regex,
                input=input_string,
            ))
            return (obj, match_obj)
    return (None, None)



from django.conf.urls.defaults import *

import fa.views as views

urlpatterns = patterns('',
    (r'^login$', 'django.contrib.auth.views.login', {'template_name': 'fa/login.html'}),
    (r'^logout$', 'django.contrib.auth.views.logout', {'template_name': 'fa/logout.html'}),
    url(r'^smssync$', views.smssync),
)



# -*- coding: utf-8 -*-
import datetime
from south.db import db
from south.v2 import SchemaMigration
from django.db import models


class Migration(SchemaMigration):

    def forwards(self, orm):
        # Adding model 'IncomingMessage'
        db.create_table('fa_incomingmessage', (
            ('id', self.gf('django.db.models.fields.AutoField')(primary_key=True)),
            ('sender', self.gf('django.db.models.fields.CharField')(max_length=25)),
            ('content', self.gf('django.db.models.fields.CharField')(max_length=255)),
            ('send_time', self.gf('django.db.models.fields.DateTimeField')()),
            ('transport', self.gf('django.db.models.fields.CharField')(max_length=255)),
        ))
        db.send_create_signal('fa', ['IncomingMessage'])

        # Adding model 'OutgoingMessage'
        db.create_table('fa_outgoingmessage', (
            ('id', self.gf('django.db.models.fields.AutoField')(primary_key=True)),
            ('target', self.gf('django.db.models.fields.CharField')(max_length=20)),
            ('content', self.gf('django.db.models.fields.CharField')(max_length=255)),
            ('transport', self.gf('django.db.models.fields.CharField')(max_length=255)),
        ))
        db.send_create_signal('fa', ['OutgoingMessage'])


    def backwards(self, orm):
        # Deleting model 'IncomingMessage'
        db.delete_table('fa_incomingmessage')

        # Deleting model 'OutgoingMessage'
        db.delete_table('fa_outgoingmessage')


    models = {
        'fa.incomingmessage': {
            'Meta': {'object_name': 'IncomingMessage'},
            'content': ('django.db.models.fields.CharField', [], {'max_length': '255'}),
            'id': ('django.db.models.fields.AutoField', [], {'primary_key': 'True'}),
            'send_time': ('django.db.models.fields.DateTimeField', [], {}),
            'sender': ('django.db.models.fields.CharField', [], {'max_length': '25'}),
            'transport': ('django.db.models.fields.CharField', [], {'max_length': '255'})
        },
        'fa.outgoingmessage': {
            'Meta': {'object_name': 'OutgoingMessage'},
            'content': ('django.db.models.fields.CharField', [], {'max_length': '255'}),
            'id': ('django.db.models.fields.AutoField', [], {'primary_key': 'True'}),
            'target': ('django.db.models.fields.CharField', [], {'max_length': '20'}),
            'transport': ('django.db.models.fields.CharField', [], {'max_length': '255'})
        }
    }

    complete_apps = ['fa']


import argparse
import datetime
import httplib
import json
import urllib

def test_smssync(hostname, path, phone_number, message, secret, time=None):
    if time == None:
        time = datetime.datetime.now()
    time_str = time.strftime("%m-%d-%Y %H:%M")
    params = urllib.urlencode({
        'from': phone_number,
        'message': message,
        'secret': secret,
        'sent_timestamp': time_str,
    })
    headers = {
        'Content-type': 'application/x-www-form-urlencoded',
        'Accept': 'text/plain',
    }
    conn = httplib.HTTPConnection(hostname)
    conn.request('POST', path, params, headers)
    response = conn.getresponse()
    data = response.read()
    if response.status != 200:
        print data
        raise httplib.BadStatusLine("Server returned {code}.".format(
            code=response.status
        ))
    conn.close()
    return json.loads(data)


def main():
    parser = argparse.ArgumentParser(description='Send sample messages to SMSSync.')
    parser.add_argument('--hostname', required=True, help='Hostname of server.')
    parser.add_argument('--path', required=True, help='Path of SMSSync view.')
    parser.add_argument('--number', required=True, help='Phone number this message is from.')
    parser.add_argument('--message', required=True, help='Message to send.')
    parser.add_argument('--secret', required=True, help='Secret to use for connection.')
    args = parser.parse_args()
    print "Connecting to {hostname}{path}".format(hostname=args.hostname, path=args.path)
    print test_smssync(
        args.hostname,
        args.path,
        args.number,
        args.message,
        args.secret,
    )


if __name__ == '__main__':
    main()



from setuptools import setup


with open("README") as file:
    long_description = file.read()

setup(
    name="FoneAstra",
    version="3.0.0",
    license="BSD",
    long_description=long_description,

    install_requires = [
        "django",
        "django-nose",
        "djtables",
        "djappsettings",
        "django-celery",
        "South",
        "supervisor",
    ],

    packages = ['fa'],
    include_package_data = True,

    author="University of Washington CSE",
    author_email="pittsw@uw.edu",

    maintainer="University of Washington CSE",
    maintainer_email="pittsw@uw.edu",

    description="Django package necessary for running a FoneAstra server.",
    url="https://bitbucket.org/pittsw/foneastra-base",
    
    classifiers=[
        "Development Status :: 4 - Beta",
        "Framework :: Django",
        "Intended Audience :: Healthcare Industry",
    ],
)



from datetime import datetime
import os,time,calendar

from django import template
from django.conf import settings
from django.core.urlresolvers import reverse
import pytz

from fa.apps import FoneAstraApp


register = template.Library()


@register.assignment_tag
def find_all_fa_apps():
    return ((reverse(app.app_index), app.app_name) for app
        in FoneAstraApp.find_all_fa_apps())
        

@register.filter
def to_js_timestamp(value):
    try:
        #Use this conversion if using the flot graphs.  It needs the
        #Flot graphs expect milliseconds since the epoch in local.
        return int(calendar.timegm(value.timetuple())*1000)
    except AttributeError:
        return ''


@register.filter
def to_string(value):
    return '\'%(value)s\'' % {'value': value}
